#!/usr/bin/python
# Copyright 2016 Cumulus Networks, Inc.
#
# mroute-check asic backend helper

import re
import socket
import struct
import sys
import cumulus.porttab

try:
    import bcmshell
except:
    bcmshell = None

class PortmapEntry:
    def __init__(self, dport, logical_port, linux_port):
        self.dport = dport
        self.logical_port = logical_port
        self.linux_port = linux_port
        self.port_vid = -1

    def update_port_vid(self, port_vid):
        self.port_vid = port_vid

    def __str__(self):
        return ('dport: %s, lport: %d, linux_port: %s, port_vid: %d' %\
                    (self.dport, self.logical_port, self.linux_port, self.port_vid))

class VlanXlate:
    def __init__(self, lport, ivid, evid):
        self.lport = lport
        self.ivid = ivid
        self.evid = evid # if -1 packet is shipped untagged

    def __str__(self):
        return ('lport: %s, ivid: %d, evid: %d' %\
                (self.lport, self.ivid, self.evid))

class VlanEntry:
    def __init__(self, vid, untagged_dports):
        self.vid = vid
        self.untagged_dports = untagged_dports

    def __str__(self):
        return ('vid: %d, untagged_ports: %s' %\
                (self.vid, ','.join(self.untagged_dports)))

class Encap:
    def __init__(self, linux_port, bcm_dport, intf, vlan, evid):
        self.linux_port = linux_port
        self.bcm_dport = bcm_dport
        self.intf = intf
        self.vlan = vlan
        self.evid = evid

    def __str__(self):
        _encap_str = 'port: %s' % self.linux_port
        if self.evid != -1:
            _encap_str += ', vlan: %d' % self.evid
        return _encap_str

class ReplicationGroup:
    def __init__(self, mcast_grp, encaps):
        # encaps is a list of Encap
        self.mcast_grp = mcast_grp
        self.encaps = encaps

    def __str__(self):
        dispStr = ''
        for encap in self.encaps:
            dispStr += '    %s\n' % encap
        return dispStr

class Hmroute:
    def __init__(self, sip, dip, iif, mcast_grp):
        self.sip = sip
        self.dip = dip
        self.iif = iif
        self.grp = mcast_grp | 0x2000000

    def __str__(self):
        return ('(%s, %s), iif: %d' % (self.sip, self.dip, self.iif))

class Asic:
    def __init__(self):
        self.pt = cumulus.porttab.porttab()
        self.bcm = bcmshell.bcmshell()

        # list of PortmapEntry
        self.portmap = []
        self._parse_portmap()
        # XXX - drop parse_ports if not needed in the end
        #self._parse_ports()

        # XXX - parsing l3 intf table can be expensive need to convert this
        # into an on-demand-cache
        # key - intf_id, info - vlan
        self.l3_intfs = {}
        self._parse_l3_intf()

        # list of VlanXlate
        self.vlan_xlate = []
        self._parse_vlan_xlate()

        # key: vlan, info: vlan_entry; this is an on-demand cache i.e. not the
        # entire vlan table
        self.vlans = {}

        # key - mcast_grp, info - ReplicationGroup
        self.replication_groups = {}
        self._parse_multicast_repln_groups()

        # key - '(a.b.c.d, a.b.c.d)', info - Hmroute
        self.mroutes = {}
        self._parse_l3_entry_ipmc_table()

    def __str__(self):
        ipmr_str = 'Asic ip mroutes:\n'
        for key in self.mroutes:
            ipmr_str += self._dump_one_mroute(self.mroutes[key])
        return ipmr_str

    def _parse_portmap(self):
        # Sample "show portmap" output (T2) -
        #
        #             logical  physical    mmu   ucast Qbase  mcast Qbase
        #      cpu0      0         0        52         0         2000
        #       xe0      1        25         6        72         1540
        # >>>>>>>>>>>>>>>>>>>>>>>>>> SNIP >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        try:
            portmap_str = self.bcm.run('show portmap')
        except IOError:
            # bcmshell is not accessible, maybe switchd is not running?
            raise RuntimeError('show portmap timed out')

        portmap_lines = portmap_str.splitlines()
        if len(portmap_lines) < 2 or 'Error' in portmap_lines[0]:
            raise RuntimeError('show portmap is not supported on this chip')

        try:
            logicalIndex = portmap_lines[0].split().index('logical') + 1
        except ValueError:
            raise RuntimeError('show portmap format unexpected')

        for port_str in portmap_lines[1:]:
            port_cols = port_str.split()
            dport = port_cols[0]
            logical_port = int(port_cols[logicalIndex])
            try:
                linux_port = self.pt.bcm2linux(dport)
            except:
                linux_port = None
            portmap_entry = PortmapEntry(dport, logical_port, linux_port)
            self.portmap.append(portmap_entry)

    def _parse_ports(self):
        try:
            dispStr = self.bcm.run('dump chg port')
        except IOError:
            raise RuntimeError('dump chg port')
        entries = dispStr.splitlines()
        for entry in entries:
            key, fields = self._field_dict_build(entry)
            if not key:
                continue
            port_vid = fields.get('PORT_VID', '0x0')
            if '0x' in port_vid:
                port_vid = int(port_vid, 16)
            else:
                port_vid = int(port_vid)
            self._update_portmap_port_vid(int(key), port_vid)

    def _parse_vlan(self, vid):
        cmd = 'vlan show %d' % vid
        try:
            dispStr = self.bcm.run(cmd)
        except IOError:
            raise RuntimeError(cmd)
        pat = re.compile('vlan (?P<vlan>\d+).*untagged (?P<ports>\S+)')
        obj = pat.search(dispStr)
        if obj:
            vlan = int(obj.group('vlan'))
            port_str = obj.group('ports')
            cmd = 'ports ' + port_str
            try:
                dispStr = self.bcm.run(cmd)
            except IOError:
                raise RuntimeError(cmd)
            portList = []
            ports = dispStr.splitlines()
            if len(ports) < 3:
                return

            for port in ports[2:]:
                portList.append(port.split()[0].replace('(', ''))
            self.vlans[vlan] = VlanEntry(vlan, portList)

    def _parse_l3_intf(self):
        # Sample output:
        # Free L3INTF entries: 8153
        # Unit  Intf  VRF Group VLAN    Source Mac     MTU TTL Tunnel InnerVlan  NATRealm
        # ------------------------------------------------------------------
        # 0     1     0     0     3001 44:38:39:00:69:84  1500 0    0     0     0
        try:
            dispStr = self.bcm.run('l3 intf show')
        except IOError:
            raise RuntimeError('l3 intf show')

        lines = dispStr.splitlines()
        title = lines[1].split()
        intf_index = title.index('Intf')
        vlan_index = title.index('VLAN')

        for line in lines[3:]:
            l3_intf = line.split()
            if len(l3_intf) < vlan_index:
                continue
            intf = int(l3_intf[intf_index])
            vlan = int(l3_intf[vlan_index])
            self.l3_intfs[intf] = vlan

    def _parse_ipmc_table(self):
        try:
            dispStr = self.bcm.run('ipmc table show')
        except IOError:
            raise RuntimeError('ipmc table show timed out')

        # sample format -
        # SRC IP ADDRESS  MC IP ADDRESS     VLAN MOD PORT V COS TS   GROUP     VRF
        # CLASS HIT
        # 0.0.0.0         239.1.1.3          0   0    4   1  -1  0   0x2000002   0
        # 0     no
        # 33.1.1.1        239.1.1.1          0   0    5   1  -1  0   0x2000003   0
        # 0     n
        addr_str_len = 2
        lines = dispStr.splitlines()
        title = lines[0].split()
        src_index = title.index('SRC')
        mc_index = title.index('MC') - addr_str_len
        grp_index = title.index('GROUP') - (2 * addr_str_len)

        for line in lines[2:]:
            ipmc = line.split()
            if len(ipmc) < grp_index:
                continue
            sip = ipmc[src_index]
            dip = ipmc[mc_index]
            mcast_grp = int(ipmc[grp_index])
            mroute = Hmroute(sip, dip, -1, mcast_grp)
            self.mroutes['(%s, %s)' % (sip, dip)] = mroute

    def _parse_vlan_xlate(self):
        try:
            dispStr = self.bcm.run('vlan translate egress show')
        except IOError:
            raise RuntimeError('vlan translate egress show')
        entries = dispStr.splitlines()
        for entry in entries:
            fields = entry.split(',')
            tvs = {}
            for field in fields:
                if 'modid' in field:
                    pat = re.compile('modid = (?P<modid>\d+) port = (?P<port>\d+)')
                    obj = pat.search(field)
                    if obj:
                        tvs['modid'] = obj.group('modid')
                        tvs['port'] = obj.group('port')
                else:
                    tv = field.strip().split('=')
                    tvs[tv[0].strip()] = tv[1].strip()

            # XXX - temporarily assume that modPort is the same as logical
            # port
            port = int(tvs.get('port', '-1'))
            old_vid = int(tvs.get('old_vid', '-1'))
            new_vid = int(tvs.get('new_vid', '-1'))
            self.vlan_xlate.append(VlanXlate(port, old_vid, new_vid))

    def _parse_l3_entry_ipmc_table(self):
        try:
            dispStr = self.bcm.run('dump chg l3_entry_ipv4_multicast')
        except IOError:
            raise RuntimeError('dump chg l3_entry_ipv4_multicast')
        entries = dispStr.splitlines()
        for entry in entries:
            key, fields = self._field_dict_build(entry)
            if not key:
                continue
            addr = int(fields.get('SOURCE_IP_ADDR', '0x0'), 16)
            sip = socket.inet_ntoa(struct.pack('!L', addr))
            addr = int(fields['GROUP_IP_ADDR'], 16)
            dip = socket.inet_ntoa(struct.pack('!L', addr))
            iif = fields['EXPECTED_L3_IIF']
            iif = int(iif, 16) if '0x' in iif else int(iif)
            mcast_grp = int(fields['L3MC_INDEX'])
            mroute = Hmroute(sip, dip, iif, mcast_grp)
            self.mroutes['(%s, %s)' % (sip, dip)] = mroute

    def _parse_one_multicast_repln_group(self, dispStr):
        group = dispStr.splitlines()
        if len(group) < 1 or 'L2' in group[0]:
            return

        group_id = int(group[0].split()[0], 16)
        encaps = []
        for encap in group[1:]:
            pat = re.compile('\s+port (?P<port>\S+), encap id (?P<intf>\d+)')
            obj = pat.match(encap)
            if obj:
                dport = obj.group('port')
                intf = int(obj.group('intf'))
                vlan = self.l3_intfs.get(intf, -1)
                evid = self._ivid2evid(dport, vlan)
                encaps.append(Encap(self.pt.bcm2linux(dport), dport, intf, vlan, evid))
        repln_group = ReplicationGroup(group_id, encaps)
        self.replication_groups[group_id] = repln_group

    def _parse_multicast_repln_groups(self):
        try:
            dispStr = self.bcm.run('multicast show')
        except IOError:
            raise RuntimeError('multicast show')
        entries = dispStr.split('Group')[1:]
        for entry in entries:
            self._parse_one_multicast_repln_group(entry)

    def _field_dict_build(self, entry):
        key = None
        fields = {}

        # format table[index]: <field1=val1,field2=val2>
        pat=re.compile('\[(?P<key>\d+)\]: \<(?P<data>\S+)\>')
        obj = pat.search(entry)
        if not obj:
            return key, fields

        data = obj.group('data')
        key = obj.group('key')

        for entry in data.split(','):
            entry=entry.split('=')
            if len(entry) > 1:
                fields[entry[0]]=entry[1]
        return key, fields

    def _update_portmap_port_vid(self, logical_port, port_vid):
        for pe in self.portmap:
            if pe.logical_port == logical_port:
                pe.update_port_vid(port_vid)

    def _dport2logical_port(self, dport):
        for pe in self.portmap:
            if pe.dport == dport:
                return pe.logical_port
        return -1

    def _is_port_untagged_for_vid(self, dport, vid):
        vlan_entry = self.vlans.get(vid)

        if not vlan_entry:
            self._parse_vlan(vid)
            vlan_entry = self.vlans.get(vid)

        if not vlan_entry:
            return False

        if dport in vlan_entry.untagged_dports:
            return True

        return False

    def _ivid2evid(self, dport, ivid):
        lport = self._dport2logical_port(dport)
        if lport == -1:
            return ivid

        for entry in self.vlan_xlate:
            if entry.lport == lport and entry.ivid == ivid:
                return entry.evid

        # no vlan xlate - check if port is an untagged member of ivid
        if self._is_port_untagged_for_vid(dport, ivid):
            return -1

        return ivid

    def _dump_one_mroute(self, mroute):
        mroute_str = '%s\n' % mroute
        replication_group = self.replication_groups.get(mroute.grp)
        if replication_group:
            mroute_str += '%s' % replication_group
        return mroute_str

    def mroute_eq(self, kmroute):
        # return True if entries match
        # XXX - need to expand the kernel entry out into l2 and l3 encaps
        return True

    def get_mroutes(self):
        # return mroute keys as a set
        return set(self.mroutes)

    def dump_mroutes(self, filter_mroutes=None):
        if not filter_mroutes:
            for key in self.mroutes:
                sys.stdout.write('%s\n' % self._dump_one_mroute(self.mroutes[key]))
            return

        for key in filter_mroutes:
            mroute = self.mroutes.get(key)
            if mroute:
                sys.stdout.write('%s\n' % self._dump_one_mroute(mroute))
